// Various light patterns for the RGB light string
// Some patterns are best for a 150 bit light string
//
// There are two identical versions for all the routines:
// 1) Uses the default blocking delay() function. See example WS2812_sample.ino
// 2) Uses a user-defined nonblockingDelay() callback function. See example WS2812_sample2.ino
//       This user-defined function allows the inclusion of WiFi service calls during a non-blocking delay
//
// All functions have a second version without the nonblockingDelay.
// List of functions:
// Wheel(i);
// selectColor(i);
// solid(leds, nonblockingDelay, color, startPixel, endPixel, times);
// allOn( nonblockingDelay, color, times);
// wipe(leds, nonblockingDelay, color, wait, times, segments);
// rainbow(leds, nonblockingDelay, wait, times);
// movingBlocks(leds, nonblockingDelay, wait, times);
// theaterChase(leds, nonblockingDelay, color, wait, times);
// theaterChaseRainbow(leds, nonblockingDelay, wait, times);
// sparkles(leds, nonblockingDelay, numberOfLeds, wait, times);
// discoLights(leds, nonblockingDelay, blockSize, wait, times);
// moveIn(leds, nonblockingDelay, blockSize, wait, times);
// flash(leds, nonblockingDelay, color, wait, times);
// fadeIn(leds, nonblockingDelay, color, wait, maxBright);
// fadeOut(leds, nonblockingDelay, color, wait, maxBright);
// fade(leds, nonblockingDelay, color, wait, times, maxBright);
// fadingRainbow(leds, nonblockingDelay, color, wait, times);
// cyclone(leds, nonblockingDelay, color, wait, times, segments);
// range(leds, nonblockingDelay, color, startPixel, endPixel, wait);
// single(leds, nonblockingDelay, color, wait, times);
//
// Copyright 2019, 2023 Enoch Hwang


// Input a value 0 to 255 to get a color value.
// best if in steps of 32
// The colors transition from r - g - b - back to r.
// orange, yellow, green, cyan, blue, magenta, pink, red
//                                                   255               
uint32_t Wheel(byte WheelPos) {
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) {
//    Serial.print(255 - WheelPos * 3, HEX);
//    Serial.print(", 0, ");
//    Serial.println(WheelPos * 3, HEX);
    return Color(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if(WheelPos < 170) {
    WheelPos -= 85;
//    Serial.print("0, ");
//    Serial.print(WheelPos * 3, HEX);
//    Serial.print(", ");
//    Serial.println(255 - WheelPos * 3, HEX);
    return Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
    WheelPos -= 170;
//    Serial.print(WheelPos * 3, HEX);
//    Serial.print(", ");
//    Serial.print(255 - WheelPos * 3, HEX);
//    Serial.println(", 0");
  return Color(WheelPos * 3, 255 - WheelPos * 3, 0);
}

void testWheel(WS2812 &leds) {
  uint32_t c;
  for (int i=0; i<=256; i=i+16) {
//  Serial.print(i);
//  Serial.print(": ");
    c = Wheel(i & 255);
    leds.setPixelColor(100, c);
    leds.setPixelColor(101, c);
    leds.setPixelColor(102, c);
    leds.show();
    delay(2000);
  }
  
}

uint32_t selectColor(int n, int brightness=100) {
  // if (brightness > 100) brightness = 100;
  switch (n) {
  case 1: return Color(brightness, brightness, 0);          // yellow
  case 2: return Color(0, 0, brightness);                   // blue
  case 3: return Color(brightness, 0, brightness);          // magenta
  case 4: return Color(0, brightness, 0);                   // green
  case 5: return Color(0, brightness, brightness);          // cyan
  case 6: return Color(brightness, 0, 0);                   // red
  case 7: return Color(brightness, brightness, brightness); // white
  default: return Color(0, 0, 0);
  }
}


void allOff() {
  leds.clear();
}


///////////////////////////////////////////////////////////////////////////////////////////
// To allow early termination of the loops
// set exitNow = true to exit immediately
bool exitNow = false;

///////////////////////////////////////////////////////////////////////////////////////////
// the following routines use the default blocking delay() function
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////

void solid(WS2812 &leds, uint32_t color=0, int startPixel=0, int endPixel=0, int onTime=15000) {
  if (color == 0) {
    color = Wheel(random(0, 256));
  }
  if (startPixel==0 && endPixel==0) {
  endPixel = leds.numPixels();  
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  exitNow = false;
  for(int i=startPixel; i<endPixel; i++) {
    leds.setPixelColor(i, color);
  }
  leds.show();
  while (true) {
    if ((millis() - startTime > onTime) || exitNow) return;
  }
}

void allOn(WS2812 &leds, uint32_t color=0, int onTime=15000) {
  solid(leds, color, 0,0, onTime);
}


// Fill the dots one after the other
void wipe(WS2812 &leds, uint32_t color=0, int wait=5, int onTime=15000, int segments=1) {
  if (wait <= 0) wait = 5;
  if (onTime==0) onTime = 15000;
  unsigned long startTime = millis();
  int segmentLength;
  if (leds.numPixels() <= 30) segments = 1;
  if (segments == 1) {
    segmentLength = leds.numPixels();
  } else {
    segmentLength = leds.numPixels() / segments;
  }
  if (leds.numPixels() <= 30 && wait == 5) {
    wait = 100; // slower for fewer leds
  } else if (leds.numPixels() <= 60 && wait == 5) {
    wait = 15;  // slower for fewer leds
  }
  uint32_t c = color;
  exitNow = false;
  while (true) {
    if (color==0) {
      c = Wheel(random(0, 256));
    }
    // up direction
    for (int i=0; i<segmentLength; i++) {
      for (int j=0; j<segments; j++) {
        leds.setPixelColor(j*segmentLength+i, c);
        leds.show();
        delay(wait);
        if (exitNow) return;
      }
    }
    // down direction
    if (color==0) {
      c = Wheel(random(0, 256));
    } else {
      c = 0;
    }
    for (int i=segmentLength-1; i>=0; i--) {
      for (int j=0; j<segments; j++) {
        leds.setPixelColor(j*segmentLength+i, c);
        leds.show();
        delay(wait);
        if (exitNow) return;
      }
    }
    if ((millis() - startTime > onTime) || exitNow) return;
  }
}

void rainbow(WS2812 &leds, int wait=5, int times=8) {
  if (wait <= 0) {
    wait = 5;
  }
  if (times <= 0) {
    times = 8;
  }
  uint16_t i, j;
  exitNow = false;
  for(j=0; j<256*times; j++) {
    for(i=0; i<leds.numPixels(); i++) {
      leds.setPixelColor(i, Wheel((i+j) & 255));
    }
    leds.show();
    delay(wait);
    if (exitNow) return;
  }
}

// works best for a 150 leds light string
// colorNumber 0: multi-color; 1 to 7: single color
void movingBlocks(WS2812 &leds, int colorNumber=0, int wait=50, int onTime=15000) {
  int c = colorNumber;
  if (wait <= 0) {
    wait = 50;
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int blockSize;
  if (leds.numPixels() <= 30) {
    blockSize = 5;
  } else {
    blockSize = 15;
  }
  uint8_t brightness;
  exitNow = false;
  delay(100);  // need this to prevent pause at start
  while (true) {
    // up direction
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j<4) brightness = 0;
          else if (j < 7) brightness = 10;
          else if (j < 10) brightness = 40;
          else brightness = 255;
          if (colorNumber == 0) c = i%7+1;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(wait);
    }
    
    // down direction
    for (int k=leds.numPixels()-1; k >= 0; k--) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j >= blockSize-4) brightness = 0;
          else if (j >= blockSize-7) brightness = 10;
          else if (j >= blockSize-10) brightness = 40;
          else brightness = 255;
          if (colorNumber == 0) c = i%7+1;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(wait);
    }
    
  }
}


// works best for a 150 leds light string
void movingBlocks2(WS2812 &leds, int wait=50, int onTime=15000) {
  int c, c1, c2;
  c1=0;
  c2=1;
  if (wait <= 0) {
    wait = 50;
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int blockSize = 15;
  uint8_t brightness = 255;
  exitNow = false;
  delay(100);  // need this to prevent pause at start
  while (true) {
    // up direction
    c1 = ++c1%7+1;
    c2 = ++c2%7+1;
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j<3 || j>=blockSize-3) c = 0;
          else if (j<5 || j>=blockSize-5) c = c1;
          else c = c2;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(wait);
    }
    
    // down direction
    c1 = ++c1%7+1;
    c2 = ++c2%7+1;
    for (int k=leds.numPixels()-1; k >= 0; k--) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j<3 || j>=blockSize-3) c = 0;
          else if (j<5 || j>=blockSize-5) c = c1;
          else c = c2;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(wait);
    }
    
  }
}

// works best for a 150 leds light string
void movingBlocks_original(WS2812 &leds, int wait=70, int onTime=15000) {
  if (wait <= 0) {
    wait = 20;
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int blockSize;
  if (leds.numPixels() <= 30) {
    blockSize = 5;
  } else {
    blockSize = 20;
  }
  uint8_t brightness;
  exitNow = false;
  delay(100);  // need this to prevent pause at start
  while (true) {
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize        
          if (j==0||j==1||j==blockSize-1||j==blockSize-2) brightness = 25;
          else if (j==2||j==3||j==blockSize-4||j==blockSize-3) brightness = 50;
          else brightness = 255;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(i%7, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(wait);
    }
  }
}


//Theatre-style crawling lights
void theaterChase(WS2812 &leds, uint32_t color=0, int wait=60, int times=5) {
  if (wait <= 0) {
    wait = 60;
  }
  if (times <= 0) {
    times = 5;
  }
  uint32_t c = color;
  exitNow = false;
  for (int t=0; t<times; t++) {  //do times cycles of chasing
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    for(int p=0; p<25; p++) {
      for (int q=0; q<3; q++) {
        for (int i=0; i<leds.numPixels(); i=i+3) {
          leds.setPixelColor(i+q, c);    //turn every third pixel on
        }
        leds.show();
        delay(wait);
        if (exitNow) return;

        for (int i=0; i<leds.numPixels(); i=i+3) {
          leds.setPixelColor(i+q, 0);        //turn every third pixel off
        }
        if (exitNow) return;
      }
    }
  }
}

//Theatre-style crawling lights with rainbow effect
void theaterChaseRainbow(WS2812 &leds, int wait=60, int times=2) {
  if (wait <= 0) {
    wait = 60;
  }
  if (times <= 0) {
    times = 5;
  }
  exitNow = false;
  for (int t=0; t<times; t++) {
    for (int j=0; j < 256; j+=5) {     // cycle all 256 colors in the wheel
      for (int q=0; q < 3; q++) {
        for (int i=0; i < leds.numPixels(); i=i+3) {
          leds.setPixelColor(i+q, Wheel((i+j) & 255));    //turn every third pixel on
        }
        leds.show();
        delay(wait);
        if (exitNow) return;

        for (int i=0; i < leds.numPixels(); i=i+3) {
          leds.setPixelColor(i+q, 0);        //turn every third pixel off
        }
        if (exitNow) return;
      }
    }
  }
}

void sparkles(WS2812 &leds, int onTime=15000) {
  int numberOfLeds = leds.numPixels()/4;
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int p;
  uint32_t c;
  exitNow = false;
  delay(100);  // need this to prevent pause at start
  while (true) {
    for (int i=0; i<numberOfLeds; i++) {
      p = random(0, leds.numPixels());
      c = Wheel(random(0,255));
      leds.setPixelColor(p, c);
      if ((millis() - startTime > onTime) || exitNow) return;
      yield();
    }
    leds.show();
    delay(100);
    leds.clear();
  }
}

void sequence(WS2812 &leds, int onTime=15000) {
  int c=0;
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  exitNow = false;
  delay(100);  // need this to prevent pause at start
  while (true) {
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+14) { // one time for all leds
        for (int j=0; j < 14; j+=2) {  // blocks of blockSize
          c = (c+1)%7+1;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, 255));
          leds.setPixelColor((i+j+k+1)%leds.numPixels(), 0);
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(500);
    }
  }   
}

void discoLights(WS2812 &leds, int blockSize=30, int wait=100, int times=100) {
  if (blockSize <= 0) {
    blockSize = 30;
  }
  if (wait <= 0) {
    wait = 100;
  }
  if (times <= 0) {
    times = 100;
  }
  uint32_t c;
  exitNow = false;
  for (int t=0; t<times; t++) {
//    for (int i=0; i<leds.numPixels(); i=i+n) {
    for (int i=0; i<leds.numPixels(); i=i+blockSize) {
      //n = random(4, 10) * 8;  // block size
      c = Wheel(random(0,255)); // color
//      for (int j=i; j<i+n; j++) {
      for (int j=i; j<i+blockSize; j++) {
        leds.setPixelColor(j, c);
    if (exitNow) return;
      }
    }
  leds.show();
  delay(wait);
  }
  allOff();
}

// works best for a 150 leds light string
void moveIn(WS2812 &leds, int blockSize=30, int wait=100, int times=14) {
  if (blockSize <= 0) {
    blockSize = 30;
  }
  if (wait <= 0) {
    wait = 100;
  }
  if (times <= 0) {
    times = 14;
  }
  uint32_t c;
  int numberOfLeds = leds.numPixels();
  int segments = leds.numPixels() / blockSize;
  exitNow = false;
  for (int t=0; t<times; t++) {
    c = selectColor(t%7);
    for (int i=0; i<=blockSize/2; i++) {
      for (int j=0; j<numberOfLeds; j+=blockSize) {
        leds.setPixelColor(i+j, c);
        leds.setPixelColor(j+blockSize-i, c);
        if (exitNow) return;
      }
      leds.show();
      delay(wait);
    }
  }
}

// flash the leds
void flash(WS2812 &leds, uint32_t color=0, int wait=1000, int times=10) {
  if (wait <= 0) {
    wait = 1000;
  }
  if (times <= 0) {
    times = 10;
  }
  uint32_t c = color;
  exitNow = false;
  for (int t=0; t<times; t++) {
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    allOn(leds, c, wait/2);
    if (exitNow) return;
    allOff();
    delay(wait/2);
    if (exitNow) return;
  }
}

// fade the leds in
void fadeIn(WS2812 &leds, uint32_t color=0, int wait=200, int maxBright=100) {
  if (wait <= 0) {
    wait = 200;
  }
  uint32_t c = color;
  if (color == 0) {
    color = Wheel(random(0, 255));
  }
  // fade in
  exitNow = false;
  for (int i=0; i<maxBright; i++) {  // don't go to full bright because not enough power
    leds.setBrightness(i);  // 0=off; 255=full bright
    allOn(leds, color, wait);
    // if (exitNow) {
      // leds.setBrightness(maxBright); // restore default brightness
      // return;
    // }
  }
  // leds.setBrightness(maxBright);
}

// fade the leds out
void fadeOut(WS2812 &leds, uint32_t color=0, int wait=200, int maxBright=100) {
  if (wait <= 0) {
    wait = 200;
  }
  if (color == 0) {
    color = Wheel(random(0, 255));
  }
  // fade out
  exitNow = false;
  for (int i=maxBright; i>=0; i--) {
    leds.setBrightness(i);  // 0=off; 255=full bright
    allOn(leds, color, wait);
    // if (exitNow) {
      // leds.setBrightness(maxBright); // restore default brightness
      // return;
    // }
  }
  allOff();
  leds.setBrightness(maxBright);
}


void fade(WS2812 &leds, uint32_t color=0, int wait=6000, int times=8, int maxBright=100) {
  if (wait <= 0) {
    wait = 6000;
  }
  if (times <= 0) {
    times = 8;
  }
  int w = wait/512;
  uint32_t c = color;
  allOff();
  for (int t=0; t<times; t++) {
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    // fade in
    for (int i=0; i<maxBright; i++) {  // don't go to full bright because not enough power
      leds.setBrightness(i);  // 0=off; 255=full bright
      allOn(leds, c, w);
      //#ifdef EXITNOW  // this flag allows early exit of loop
        if (exitNow) {
          leds.setBrightness(maxBright);  // restore default brightness
          return;
        }
      //#endif
    }
    // fade out
    for (int i=maxBright; i>=0; i--) {
      leds.setBrightness(i);  // 0=off; 255=full bright
      allOn(leds, c, w);
      //#ifdef EXITNOW  // this flag allows early exit of loop
        if (exitNow) {
          leds.setBrightness(maxBright);  // restore default brightness
          return;
        }
      //#endif
    }
    delay(wait/20);
  }
  leds.setBrightness(maxBright);
}

// fade through all the colors
// returns the ending color
int fadingRainbow(WS2812 &leds, int color=0, int wait=80, int times=1) {
  if (wait <= 0) {
    wait = 40;
  }
  if (times <= 0) {
    times = 1;
  }
  if (color == 0) {
    color = random(0, 255); // pick a random starting color
  }
  exitNow = false;
  for (int t=0; t<times; t++) {
    for (int c=0; c<256; c++) {
      allOn(leds, Wheel(color), wait);
      color++;
      if (color >= 255) color = 0;
      if (exitNow) {
        return color;
      }
    }
  }
  return color;
}


void cyclone(WS2812 &leds, uint32_t color=0, int wait=0, int onTime=15000, int segments=1) {
  uint32_t c = color;
  uint8_t r, g, b;

  // set different end pixel and wait time for different RGB lengths
  int endPixel, w = wait;
  if (leds.numPixels() <= 8) {
    endPixel = 2;
    if (wait == 0) w = 60;
  } else if (leds.numPixels() <= 30) {
    endPixel = 3;
    if (wait == 0) w = 30;
  } else {
    endPixel = 12;
    if (wait == 0) w = 5;
  }

  if (onTime==0) onTime = 15000;
  unsigned long startTime = millis();

  int segmentLength;
  if (leds.numPixels() <= 60) segments = 1;
  if (segments == 1) {
    segmentLength = leds.numPixels();
  } else {
    segmentLength = leds.numPixels() / segments;
  }

  exitNow = false;
  while (true) {
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    unpackColor(c, r,g,b);
    uint8_t r2,g2,b2,r3,g3,b3,r4,g4,b4,r5,g5,b5;
    // set different brightness for the three colors
    r2 = r >> 1;
    r3 = r >> 2;
    r4 = r >> 3;
    r5 = r >> 4;
    g2 = g >> 1;
    g3 = g >> 2;
    g4 = g >> 3;
    g5 = g >> 4;
    b2 = b >> 1;
    b3 = b >> 2;
    b4 = b >> 3;
    b5 = b >> 4;

    // up direction
    for(int i=0; i<leds.numPixels()+endPixel; i++) {
      if (leds.numPixels() <= 8) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i-1, r4,g4,b4);
        leds.setPixelColor(i-2, 0,0,0);
      } else if (leds.numPixels() <= 30) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i-1, r3,g3,b3);
        leds.setPixelColor(i-2, r5,g5,b5);
        leds.setPixelColor(i-3, 0,0,0);
      } else {
        for (int j=0; j<segments; j++) {
          leds.setPixelColor(j*segmentLength+i, r,g,b);
          leds.setPixelColor(j*segmentLength+i-1, r,g,b);
          leds.setPixelColor(j*segmentLength+i-2, r,g,b);
          leds.setPixelColor(j*segmentLength+i-3, r,g,b);
          leds.setPixelColor(j*segmentLength+i-4, r,g,b);
          leds.setPixelColor(j*segmentLength+i-5, r,g,b);
          leds.setPixelColor(j*segmentLength+i-6, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i-7, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i-8, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i-9, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i-10, r4,g4,b4);
          leds.setPixelColor(j*segmentLength+i-11, r5,g5,b5);
          leds.setPixelColor(j*segmentLength+i-12, 0,0,0);
        }
      }
      leds.show();
      // slows down when reaching the end
      if (i < leds.numPixels()*.60) {
        delay(w);
      } else if (i < leds.numPixels()*.80) {
        delay(w<<1);
      } else if (i < leds.numPixels()*.90) {
        delay(w<<2);
      } else {
        delay(w<<3);
      }
      //#ifdef EXITNOW  // this flag allows early exit of loop
      if (exitNow) return;
      //#endif    
    }
    // down direction
    for(int i=leds.numPixels(); i>=-endPixel; i--) {
      if (leds.numPixels() <= 8) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i+1, r4,g4,b4);
        leds.setPixelColor(i+2, 0,0,0);
      } else if (leds.numPixels() <= 30) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i+1, r3,g3,b3);
        leds.setPixelColor(i+2, r5,g5,b5);
        leds.setPixelColor(i+3, 0,0,0);
      } else {
        for (int j=0; j<segments; j++) {
          leds.setPixelColor(j*segmentLength+i, r,g,b);
          leds.setPixelColor(j*segmentLength+i+1, r,g,b);
          leds.setPixelColor(j*segmentLength+i+2, r,g,b);
          leds.setPixelColor(j*segmentLength+i+3, r,g,b);
          leds.setPixelColor(j*segmentLength+i+4, r,g,b);
          leds.setPixelColor(j*segmentLength+i+5, r,g,b);
          leds.setPixelColor(j*segmentLength+i+6, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i+7, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i+8, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i+9, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i+10, r4,g4,b4);
          leds.setPixelColor(j*segmentLength+i+11, r5,g5,b5);
          leds.setPixelColor(j*segmentLength+i+12, 0,0,0);
        }
      }
      leds.show();
      // slows down when reaching the end
      if (i < leds.numPixels()*.10) {
        delay(w<<3);
      } else if (i < leds.numPixels()*.20) {
        delay(w<<2);
      } else if (i < leds.numPixels()*.40) {
        delay(w<<1);
      } else {
        delay(w);
      }
      if ((millis() - startTime > onTime) || exitNow) return;
    }
  }
}

void range(WS2812 &leds, uint32_t color=0, int startPixel=0, int endPixel=0, int wait=0) {
  if (endPixel == 0) {
    endPixel = leds.numPixels();
  }
  if (endPixel < startPixel) {
    startPixel = 0;
    endPixel = leds.numPixels();
  }
  if (color == 0) {
    color = Wheel(random(0, 256));
  }
    for(int i=startPixel; i<endPixel; i++) {
      leds.setPixelColor(i, color);
    }
    leds.show();
}

// each time this function is called, it will turn on the next led and turn off the previous led
void single(WS2812 &leds, uint32_t color=0, int wait=0, int times=1) {
  if (wait <= 0) {
    if (leds.numPixels() > 60) {
      wait = 50;
    } else if (leds.numPixels() > 12) {
      wait = 100;
    } else {
      wait = 1000;
    }
  }
  if (times <= 0) {
    times = leds.numPixels();
  }
  static int p=0, q=leds.numPixels()-1;
  uint32_t c = color;
  exitNow = false;
  for (int t=0; t<times; t++) {
    if (color == 0) c = Wheel(random(0,255));
    leds.setPixelColor(p, c); // turn on pixel
    leds.setPixelColor(q, 0); // turn off previous pixel
    q=p;
    p++;
    if (p == leds.numPixels()) p = 0;
    leds.show();
    delay(wait);
    if (exitNow) return;
  }
}


///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
// use the user-defined non-blocking nonblockingDelay() callback function
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////

void solid(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int startPixel=0, int endPixel=0, int onTime=15000) {
  if (color == 0) {
    color = Wheel(random(0, 256));
  }
  if (startPixel==0 && endPixel==0) {
  endPixel = leds.numPixels();  
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  exitNow = false;
  for(int i=startPixel; i<endPixel; i++) {
    leds.setPixelColor(i, color);
  }
  leds.show();
  while (true) {
    nonblockingDelay(1);
    if ((millis() - startTime > onTime) || exitNow) return;
  }
}


void allOn(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int onTime=20000) {
  solid(leds, nonblockingDelay, color, 0,0, onTime);
}


void rainbow(WS2812 &leds, void (*nonblockingDelay)(int), int onTime=15000, int wait=30) {
  unsigned long startTime = millis();
  int j = 0;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    for(int i=0; i<leds.numPixels(); i++) {
      leds.setPixelColor(i, Wheel((i+j)));
      if (exitNow) break;
      yield();
    }
    leds.show();
    j++;
    nonblockingDelay(wait);
    if ((millis() - startTime > onTime) || exitNow) return;
  }
  startTime = 0;
}

// works best for a 150 leds light string
// colorNumber 0: multi-color; 1 to 7: single color
void movingBlocks(WS2812 &leds, void (*nonblockingDelay)(int), int colorNumber=0, int wait=50, int onTime=15000) {
  int c = colorNumber;
  if (wait <= 0) {
    wait = 50;
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int blockSize;
  if (leds.numPixels() <= 30) {
    blockSize = 5;
  } else {
    blockSize = 20;
  }
  uint8_t brightness;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    // up direction
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j<4) brightness = 0;
          else if (j < 7) brightness = 10;
          else if (j < 10) brightness = 40;
          else brightness = 255;
          if (colorNumber == 0) c = i%7+1;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      nonblockingDelay(wait);
    }
    
    // down direction
    for (int k=leds.numPixels()-1; k >= 0; k--) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j >= blockSize-4) brightness = 0;
          else if (j >= blockSize-7) brightness = 10;
          else if (j >= blockSize-10) brightness = 40;
          else brightness = 255;
          if (colorNumber == 0) c = i%7+1;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      nonblockingDelay(wait);
    }
    
  }
}

// works best for a 150 leds light string
void movingBlocks2(WS2812 &leds, void (*nonblockingDelay)(int), int wait=50, int onTime=15000) {
  int c, c1, c2;
  c1=0;
  c2=1;
  if (wait <= 0) {
    wait = 50;
  }
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int blockSize = 15;
  uint8_t brightness = 255;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    // up direction
    c1 = ++c1%7+1;
    c2 = ++c2%7+1;
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j<3 || j>=blockSize-3) c = 0;
          else if (j<5 || j>=blockSize-5) c = c1;
          else c = c2;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      nonblockingDelay(wait);
    }
    
    // down direction
    c1 = ++c1%7+1;
    c2 = ++c2%7+1;
    for (int k=leds.numPixels()-1; k >= 0; k--) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+blockSize) { // one time for all leds
        for (int j=0; j < blockSize; j++) {  // blocks of blockSize
          if (j<3 || j>=blockSize-3) c = 0;
          else if (j<5 || j>=blockSize-5) c = c1;
          else c = c2;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, brightness));
//          if (j==blockSize-1) {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), 0);
//          } else {
//            leds.setPixelColor((i+j+k)%leds.numPixels(), Wheel(i*4));
//          }
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      delay(wait);
    }
    
  }
}

//Theatre-style crawling lights
void theaterChase(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=60, int onTime=15000) {
  unsigned long startTime = millis();
  if (wait <= 0) {
    wait = 60;
  }
  uint32_t c = color;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    for(int j=0; j<256; j+=5) {
      for (int q=0; q<3; q++) {
        //turn every third pixel on
        for (int i=0; i<leds.numPixels(); i=i+3) {
          if (color == 0) c = Wheel((i+j) & 255);
          leds.setPixelColor(i+q, c);
        }
        leds.show();
        nonblockingDelay(wait);
        if ((millis() - startTime > onTime) || exitNow) return;
        //turn every third pixel off
        for (int i=0; i<leds.numPixels(); i=i+3) {
          leds.setPixelColor(i+q, 0);
        }
        if ((millis() - startTime > onTime) || exitNow) return;
      }
    }
  }
}

////Theatre-style crawling lights with rainbow effect
//void theaterChaseRainbow(void (*nonblockingDelay)(int), int wait=60, int onTime=20000) {
//  unsigned long startTime = millis();
//  if (wait <= 0) {
//    wait = 60;
//  }
//  exitNow = false;
//  while (true) {
//    for (int j=0; j < 256; j+=5) {     // cycle all 256 colors in the wheel
//      for (int q=0; q < 3; q++) {
//        for (int i=0; i < leds.numPixels(); i=i+3) {
//          leds.setPixelColor(i+q, Wheel((i+j) & 255));    //turn every third pixel on
//        }
//        leds.show();
//        nonblockingDelay(wait);
//        if ((millis() - startTime > onTime) || exitNow) return;
//        for (int i=0; i < leds.numPixels(); i=i+3) {
//          leds.setPixelColor(i+q, 0);        //turn every third pixel off
//        }
//        if ((millis() - startTime > onTime) || exitNow) return;
//      }
//    }
//  }
//}



void discoLights(WS2812 &leds, void (*nonblockingDelay)(int), int wait=200, int onTime=15000) {
  unsigned long startTime = millis();
  uint32_t c;
  int n = 0;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    for (int i=0; i<leds.numPixels(); i=i+n) {
      n = random(1, 4) * 4;  // block size
      c = Wheel(random(0,255)); // color
      for (int j=i; j<i+n; j++) {
        leds.setPixelColor(j, c);
        if ((millis() - startTime > onTime) || exitNow) return;
        yield();
      }
    }
    leds.show();
    nonblockingDelay(wait);
  }
}


void moveIn(WS2812 &leds, void (*nonblockingDelay)(int), int wait=150, int onTime=15000) {
//void moveIn(void (*nonblockingDelay)(long unsigned int), int wait=100, int onTime=20000) {
  unsigned long startTime = millis();
  uint32_t color;
  unsigned int c = random(0, 255);
  int blockSize = 12;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    color = Wheel(c);
    c = c + 96;
    for (int i=0; i<=blockSize/2; i++) {
      for (int j=0; j<leds.numPixels(); j+=blockSize) {
        leds.setPixelColor(i+j, color);
        leds.setPixelColor(j+blockSize-i, color);
        if ((millis() - startTime > onTime) || exitNow) return;
        yield();
      }
      leds.show();
//      delay(wait);
      nonblockingDelay(wait);
    }
  }
}

// flash the leds
void flash(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=1000, int onTime=15000) {
  unsigned long startTime = millis();
  if (wait <= 0) {
    wait = 1000;
  }
  uint32_t c = color;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    allOn(leds, nonblockingDelay, c, wait/2);
    allOff();
    nonblockingDelay(wait/2);
    if ((millis() - startTime > onTime) || exitNow) return;
  }
}

// fade the leds in
void fadeIn(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=200, int maxBright=100) {
  if (wait <= 0) {
    wait = 200;
  }
  uint32_t c = color;
  if (color == 0) {
    color = Wheel(random(0, 255));
  }
  // fade in
  exitNow = false;
  for (int i=0; i<maxBright; i++) {  // don't go to full bright because not enough power
    leds.setBrightness(i);  // 0=off; 255=full bright
    allOn(leds, nonblockingDelay, color, wait);
    // if (exitNow) {
      // leds.setBrightness(maxBright); // restore default brightness
      // return;
    // }
  }
  // leds.setBrightness(maxBright);
}

// fade the leds out
void fadeOut(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=200, int maxBright=100) {
  if (wait <= 0) {
    wait = 200;
  }
  if (color == 0) {
    color = Wheel(random(0, 255));
  }
  // fade out
  exitNow = false;
  for (int i=maxBright; i>=0; i--) {
    leds.setBrightness(i);  // 0=off; 255=full bright
    allOn(leds, nonblockingDelay, color, wait);
    // if (exitNow) {
      // leds.setBrightness(maxBright); // restore default brightness
      // return;
    // }
  }
  allOff();
  leds.setBrightness(maxBright);
}

void fade(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=6000, int onTime=15000) {
  unsigned long startTime = millis();
  int maxBright = leds.getBrightness();
  if (wait <= 0) {
    wait = 6000;
  }
  int w = wait/512;
  uint32_t c = color;
  allOff();
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    if (color == 0) {
      c = Wheel(random(0, 255));
    }
    // fade in
    for (int i=0; i<maxBright; i++) {  // don't go to full bright because not enough power
      leds.setBrightness(i);  // 0=off; 255=full bright
      allOn(leds, nonblockingDelay, c, w);
      if ((millis() - startTime > onTime) || exitNow) {
        leds.setBrightness(maxBright);  // restore default brightness
        return;
      }
      yield();
    }
    // fade out
    for (int i=maxBright; i>=0; i--) {
      leds.setBrightness(i);  // 0=off; 255=full bright
      allOn(leds, nonblockingDelay, c, w);
      if ((millis() - startTime > onTime) || exitNow) {
        leds.setBrightness(maxBright);  // restore default brightness
        return;
      }
      yield();
    }
    nonblockingDelay(wait/20);
  }
  leds.setBrightness(maxBright);
}

// fade through all the colors
void fadingRainbow(WS2812 &leds, void (*nonblockingDelay)(int), int wait=60, int onTime=20000) {
  unsigned long startTime = millis();
  if (wait <= 0) {
    wait = 40;
  }
  unsigned int color = random(0, 255);
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    for (int c=0; c<256; c++) {
      allOn(leds, nonblockingDelay, Wheel(color), wait);
      color++;
//      if (color >= 255) color = 0;
      if ((millis() - startTime > onTime) || exitNow) return;
    }
  }
}

int delaySpeed[] = {200,100,50,30,20,10};
void cyclone3(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=300, int times=6) {
  if (wait <= 0) {
    wait = 300;
  }
  if (times <= 0) {
    times = 6;
  }
  uint32_t c = color;
  int q=leds.numPixels()-1;
  int w;
  for (int t=0; t<times; t++) {
    if (color == 0) c = Wheel(random(0,255));
    for (int i=0; i<leds.numPixels(); i++) {
      leds.setPixelColor(i, c); // turn on pixel
      if (i == 0) q = leds.numPixels() - 1;
      else q = i - 1;
      leds.setPixelColor(q, 0); // turn off previous pixel
      leds.show();
      // w = i-(leds.numPixels()-1-i);
      // w = abs(w)*20;
      // Serial.println(w);
      if (i<2 || i>leds.numPixels()-3) w = delaySpeed[0];
      else if (i<5 || i>leds.numPixels()-6) w = delaySpeed[1];
      else if (i<8 || i>leds.numPixels()-9) w = delaySpeed[2];
      else if (i<12 || i>leds.numPixels()-13) w = delaySpeed[3];
      else if (i<16 || i>leds.numPixels()-17) w = delaySpeed[4];
      else w = delaySpeed[5];
      nonblockingDelay(w);
    }
  }
}


void cyclone(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=10, int onTime=15000) {
  unsigned long startTime = millis();
  uint32_t c = color;
  uint8_t r, g, b;
  uint8_t r2,g2,b2,r3,g3,b3,r4,g4,b4,r5,g5,b5;
  if (leds.numPixels() <= 30) wait = 30;
  int endPixel = 8, w = wait;  
  int segments = 1;
  int segmentLength = leds.numPixels() / segments;
  
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    // up direction
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    unpackColor(c, r,g,b);
    // set different brightness for the three colors
    r2 = r >> 1;
    r3 = r >> 2;
    r4 = r >> 3;
    r5 = r >> 4;
    g2 = g >> 1;
    g3 = g >> 2;
    g4 = g >> 3;
    g5 = g >> 4;
    b2 = b >> 1;
    b3 = b >> 2;
    b4 = b >> 3;
    b5 = b >> 4;
    for(int i=0; i<segmentLength+endPixel; i++) {
      if (leds.numPixels() <= 8) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i-1, r4,g4,b4);
        leds.setPixelColor(i-2, 0,0,0);
      } else if (leds.numPixels() <= 30) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i-1, r3,g3,b3);
        leds.setPixelColor(i-2, r5,g5,b5);
        leds.setPixelColor(i-3, 0,0,0);
      } else {
        for (int j=0; j<segments; j++) {
          leds.setPixelColor(j*segmentLength+i, r,g,b);
          leds.setPixelColor(j*segmentLength+i-1, r,g,b);
          leds.setPixelColor(j*segmentLength+i-2, r,g,b);
          leds.setPixelColor(j*segmentLength+i-3, r,g,b);
          leds.setPixelColor(j*segmentLength+i-4, r,g,b);
          leds.setPixelColor(j*segmentLength+i-5, r,g,b);
          leds.setPixelColor(j*segmentLength+i-6, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i-7, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i-8, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i-9, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i-10, r4,g4,b4);
          leds.setPixelColor(j*segmentLength+i-11, r5,g5,b5);
          leds.setPixelColor(j*segmentLength+i-12, 0,0,0);
        }
      }
      if (exitNow) return;
      leds.show();
      // slows down when reaching the end
      if (i < leds.numPixels()*.60) {
        nonblockingDelay(w);
      } else if (i < leds.numPixels()*.80) {
        nonblockingDelay(w<<1);
      } else if (i < leds.numPixels()*.90) {
        nonblockingDelay(w<<2);
      } else {
        nonblockingDelay(w<<3);
      }
      if (exitNow) return;
    }
    
    // down direction
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    unpackColor(c, r,g,b);
    // set different brightness for the three colors
    r2 = r >> 1;
    r3 = r >> 2;
    r4 = r >> 3;
    r5 = r >> 4;
    g2 = g >> 1;
    g3 = g >> 2;
    g4 = g >> 3;
    g5 = g >> 4;
    b2 = b >> 1;
    b3 = b >> 2;
    b4 = b >> 3;
    b5 = b >> 4;
    for(int i=segmentLength; i>=-endPixel; i--) {
      if (leds.numPixels() <= 8) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i+1, r4,g4,b4);
        leds.setPixelColor(i+2, 0,0,0);
      } else if (leds.numPixels() <= 30) {
        leds.setPixelColor(i, r,g,b);
        leds.setPixelColor(i+1, r3,g3,b3);
        leds.setPixelColor(i+2, r5,g5,b5);
        leds.setPixelColor(i+3, 0,0,0);
      } else {
        for (int j=0; j<segments; j++) {
          leds.setPixelColor(j*segmentLength+i, r,g,b);
          leds.setPixelColor(j*segmentLength+i+1, r,g,b);
          leds.setPixelColor(j*segmentLength+i+2, r,g,b);
          leds.setPixelColor(j*segmentLength+i+3, r,g,b);
          leds.setPixelColor(j*segmentLength+i+4, r,g,b);
          leds.setPixelColor(j*segmentLength+i+5, r,g,b);
          leds.setPixelColor(j*segmentLength+i+6, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i+7, r2,g2,b2);
          leds.setPixelColor(j*segmentLength+i+8, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i+9, r3,g3,b3);
          leds.setPixelColor(j*segmentLength+i+10, r4,g4,b4);
          leds.setPixelColor(j*segmentLength+i+11, r5,g5,b5);
          leds.setPixelColor(j*segmentLength+i+12, 0,0,0);
        }
      }
      if (exitNow) return;
      leds.show();
      // slows down when reaching the end
      if (i < leds.numPixels()*.10) {
        nonblockingDelay(w<<3);
      } else if (i < leds.numPixels()*.20) {
        nonblockingDelay(w<<2);
      } else if (i < leds.numPixels()*.40) {
        nonblockingDelay(w<<1);
      } else {
        nonblockingDelay(w);
      }
      if ((millis() - startTime > onTime) || exitNow) return;
    }
  }

}

void sparkles(WS2812 &leds, void (*nonblockingDelay)(int), int onTime=15000) {
  int numberOfLeds = leds.numPixels()/4;
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  int p;
  uint32_t c;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    for (int i=0; i<numberOfLeds; i++) {
      p = random(0, leds.numPixels());
      c = Wheel(random(0,255));
      leds.setPixelColor(p, c);
      if ((millis() - startTime > onTime) || exitNow) return;
      yield();
    }
    leds.show();
    nonblockingDelay(100);
    leds.clear();
  }
}

void sequence(WS2812 &leds, void (*nonblockingDelay)(int), int onTime=15000) {
  int c=0;
  if (onTime == 0) {
    onTime = 15000;
  }
  unsigned long startTime = millis();
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    for (int k=0; k < leds.numPixels(); k++) {  // make it move
      for (int i=0; i < leds.numPixels(); i=i+14) { // one time for all leds
        for (int j=0; j < 14; j+=2) {  // blocks of blockSize
          c = (c+1)%7+1;
          leds.setPixelColor((i+j+k)%leds.numPixels(), selectColor(c, 255));
          leds.setPixelColor((i+j+k+1)%leds.numPixels(), 0);
          if ((millis() - startTime > onTime) || exitNow) return;
          yield();
        }
      }
      leds.show();
      nonblockingDelay(500);
    }
  }   
}

// Fill the dots one after the other
void wipe(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int wait=5, int onTime=15000, int segments=1) {
  if (wait <= 0) wait = 5;
  if (onTime==0) onTime = 15000;
  unsigned long startTime = millis();
  int segmentLength;
  if (leds.numPixels() <= 30) segments = 1;
  if (segments == 1) {
    segmentLength = leds.numPixels();
  } else {
    segmentLength = leds.numPixels() / segments;
  }
  if (leds.numPixels() <= 30 && wait == 5) {
    wait = 100;  // slower for fewer leds
  } else if (leds.numPixels() <= 60 && wait == 5) {
    wait = 15;  // slower for fewer leds
  }
  uint32_t c = color;
  exitNow = false;
  nonblockingDelay(100);  // need this to prevent pause at start
  while (true) {
    if (color==0) {
      c = Wheel(random(0, 255));
    }
    // up direction
    for (int i=0; i<segmentLength; i++) {
      for (int j=0; j<segments; j++) {
        leds.setPixelColor(j*segmentLength+i, c);
        leds.show();
        nonblockingDelay(wait);
        if (exitNow) return;
      }
    }
    // down direction
    if (color==0) {
        c = Wheel(random(0, 255));
    } else {
        c = 0;
    }
    for (int i=segmentLength-1; i>=0; i--) {
      for (int j=0; j<segments; j++) {
        leds.setPixelColor(j*segmentLength+i, c);
        leds.show();
        nonblockingDelay(wait);
        if (exitNow) return;
      }
    }
    if ((millis() - startTime > onTime) || exitNow) return;
  }
}

void range(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0, int startPixel=0, int endPixel=0, int wait=0) {
  if (endPixel == 0) {
    endPixel = leds.numPixels();
  }
  if (endPixel < startPixel) {
    startPixel = 0;
    endPixel = leds.numPixels();
  }
  if (color == 0) {
    color = Wheel(random(0, 256));
  }
    for(int i=startPixel; i<endPixel; i++) {
      leds.setPixelColor(i, color);
    }
    leds.show();
}

// each time this function is called, it will turn on the next led and turn off the previous led
void single(WS2812 &leds, void (*nonblockingDelay)(int), uint32_t color=0) {
  static int p=0, q=leds.numPixels()-1;
  uint32_t c = color;
  if (color == 0) c = Wheel(random(0,255));
  leds.setPixelColor(p, c); // turn on pixel
  leds.setPixelColor(q, 0); // turn off previous pixel
  q=p;
  p++;
  if (p == leds.numPixels()) p = 0;
  leds.show();
}

// #endif
